# -*- coding: utf-8 -*-
### <tspGAKlasse.py>

# Mutationswahrscheinlichkeit und Selektionswahrscheinlichkeit
# sind positive ganze Zahlen zwischen 0 und 100 ( also % )

bildGroesse = 800

### Hilfsfunktionen
import random
import math
### ===== Abstand ====================================================
def abstand(stadt1,stadt2):
    # berechnet den Abstand zweier Städte
    x1,y1=stadt1
    x2,y2=stadt2
    return math.sqrt((x1-x2)**2+(y1-y2)**2)

### ===== bewerte ===================================================
def bewerte(rundtour):
    # Die Bewertungsfunktion berechnet die Summe der geometrischen
    # Abstände (Luftliniendistanz) der Rundtour.
    summe=0
    for i in range(len(rundtour)-1):
        summe+=abstand(rundtour[i],rundtour[i+1])
    summe+=abstand(rundtour[len(rundtour)-1],rundtour[0])
    return summe

### ===== besserP ===================================================
# Das Prädikat 'besserP' ist das Vergleichsprädikat zur Bewertung und
# ist daher vom speziellen Problem abhängig. Hier wird die vorgegebene
# Funktion bewerte (s.o.) genommen.
def besserP(individuum1, individuum2):
   return bewerte (individuum1) < bewerte (individuum2)

### ===== aktionP ===================================================
# Hilfsfunktion (Praedikat), die entscheidet, ob eine Aktion stattfindet
# oder nicht. Die Entscheidung hängt von der übergebenen ganzzahligen
# Wahrscheinlichkeit p ab. 0 < p < 100
def aktionP(p):
    return (random.randint(1,100) <= p)

### ===== waehle-Positionen =========================================
# Hilfsfunktion
# Wählt zwei verschiedene Zahlen aus {0 .. n-1} fuer die Position des
# Schnitts in den "Genen" aus und gib sie als Liste  zurück. 
# Die Zahlen sind stets geordnet: [klein gross]
# Unzulässige Parameter für n lösen einen Fehler aus!
def waehlePositionen(n):
    if type(n)!=int: return [0,n-1]
    if n<2: return[0,0]
    z1=random.randint(0,n-1)
    z2=random.randint(0,n-1)
    if (0 in[z1,z2]) and (n-1 in[z1,z2]):
        return waehlePositionen(n) # keine Auswahl, noch einmal versuchen
    if z1<z2: return [z1,z2]
    if z2<z1: return [z2,z1]
    return waehlePositionen(n) # gleich, noch einmal versuchen

### ===== Zufallsgenerator-setzen ====================================
# Setzt den Zufallsgenerator
# Das fuehrt bei erneutem Aufruf zur selben Folge
#random.seed(100)

### ===== Klasse TSPGA ===============================================
class TSPGA(object):

    def __init__(self):
        self.anzahlStaedte = 10
        self.staedteListe = self.erzeugeAlleStaedte(10)
        self.schrittDaten=[]

    def setzeAnzahlStaedte(self, anzahl):
        self.anzahlStaedte = anzahl

    def gibStaedte(self):
        return self.staedteListe

    def erzeugeEineStadt(self):
        # eine Stadt ist ein Zahlenpaar
        return (random.randint(1,bildGroesse-2),
                random.randint(1,bildGroesse-2))

    def erzeugeAlleStaedte(self, anzahlStaedte):
        self.staedteListe = []
        i = anzahlStaedte
        while i>0:
            stadt = self.erzeugeEineStadt()
            if not(stadt in self.staedteListe):
                self.staedteListe.append(stadt)
                i-=1
        return self.staedteListe

    def erzeugeIndividuum(self, staedteListe):
        # Erzeugt ein Individuum der Population.
        individuum = []
        staedte=[]+staedteListe # echte Kopie
        while len(staedte)>0:
            i=random.randint(1,len(staedte))-1
            individuum.append(staedte[i])
            staedte.remove(staedte[i])
        return individuum

    def erzeugePopulation(self,staedteListe,anzahl=5):
        # Erzeugt die Grundpopulation für den genetischen Prozess.
        # Parameter sind: Anzahl der Individuen und die Städteliste
        # Wert: Eine Liste von Individuen
        population = []
        for i in range(anzahl):
            population.append(self.erzeugeIndividuum(staedteListe))
        return population

    def mutiere(self, individuum):
        # Mutiert ein einzelnes Individuum.
        # Dazu wird der Teil der Tour zwischen der ersten und der zweiten
        # Position einschliesslich umgekehrt eingebaut.
        # Individuum -> Individuum
        von,bis = waehlePositionen(len(individuum))
        cut=individuum[von:bis]
        cut.reverse()
        return individuum[:von]+cut+individuum[bis:]

    def mutation(self, population, pMutatioon):
        # Steuert die Mutation für die gesamte Population.
        # Die bei der Mutation neu erzeugten Individuen werden der Population
        # hinzugefügt, dadurch ändert sich die Gesamtzahl. Deren Anpassung
        # erfolgt bei der Selektion.
        # Parameter sind: Population, Mutationswahrscheinlichkeit (pMutatioon)
        # Wert: Population zuzüglich der Mutierten
        mutierte = []
        for individuum in population:
            if aktionP(pMutatioon):
                mutierte.append(self.mutiere(individuum))
            else:
                mutierte.append(individuum)
        return population+mutierte


    def kreuze(self, individuum1, individuum2):
        # Kreuzt zwei einzelne Individuen.
        # Das Kreuzen ist beim TSP schwierig umzusetzen, da keine Staedte weg
        # fallen duerfen und umgekehrt.
        # Das Problem ist, dass eingekreuzte Abschnitte Staedte enthalten, 
        # die schon in der Resttour enthalten sind, andere aber z.T. nicht.
        # Individuum x Individuum -> (list Individuum Individuum)
        vorn,hinten = waehlePositionen(len(individuum1))
        hinten+=1 # also einschliesslich schneiden
        cut1=individuum1[vorn:hinten]
        cut2=individuum2[vorn:hinten]
        neu1 = []
        neu2 = []
        dif1 = []
        dif2 = []
        for stadt in cut1:
            if not(stadt in cut2):
                dif1.append(stadt)
        for stadt in cut2:
            if not(stadt in cut1):
                dif2.append(stadt)
        
        for stadt in individuum1[:vorn]:
            if not(stadt in cut2):
                neu1.append(stadt)
            else:
                neu1.append(dif1.pop(0))
        neu1 += cut2
        for stadt in individuum1[hinten:]:
            if not(stadt in cut2):
                neu1.append(stadt)
            else:
                neu1.append(dif1.pop(0))
        
        for stadt in individuum2[:vorn]:
            if not(stadt in cut1):
                neu2.append(stadt)
            else:
                neu2.append(dif2.pop(0))
        neu2 += cut1
        for stadt in individuum2[hinten:]:
            if not(stadt in cut1):
                neu2.append(stadt)
            else:
                neu2.append(dif2.pop(0))
        return [neu1, neu2]

    def kreuzen(self, population, pCrossingOver):
        # Steuert das Kreuzen für die gesamte Population.
        # Da beim Kreuzen neue Individuen erzeugt und der Population hinzu-
        # gefügt werden, ändert sich die Gesamtzahl. Deren Anpassung erfolgt
        # bei der Selektion.
        # Die zu kreuzenden Individuen werden nach dem Zufallsprinzip ausge-
        # wählt.
        # Parameter sind: Population, crossing-over-Wahrscheinlichkeit
        # Wert:  Population zuzüglich der Gekreuzten
        gekreuzte = []
        n=len(population)
        for individuum in population:
            if aktionP(pCrossingOver):
                anderes = population[random.randint(0,n-1)]
                while anderes == individuum:
                    anderes = population[random.randint(0,n-1)]
                gekreuzte+=self.kreuze(individuum,anderes)
        return population+gekreuzte

    def selektiere(self, individuum1, individuum2):
        # Selektiert von zwei Individuen das mit der höheren Bewertungszahl.
        if besserP(individuum1, individuum2):
            return individuum1
        return individuum2

    def selektion(self, population, gesamtzahl):
        # Es gibt viele veschiedene Selektionsmechanismen. Beispielsweise
        # kann die Auswahl proportional zur Bewertungsfunktion erfolgen.
        # Hier werden die vorhandenen Individuen einfach zufällig ausgewählt
        # und einer paarweisen Auswahl unterzogen: Selektiert wird aus der
        # Menge der Individuen durch direkten Vergleich von je zwei
        # Individuen. Der Parameter Gesamtzahl gewährleistet, dass wieder
        # die ursprüngliche Populationsgrösse hergestellt wird.
        # Dazu wird mit der Selektion abgebrochen, wenn eine genügend grosse
        # Zahl (-> Gesamtzahl) ausgewählt wurde.
        # Am Beginn wird die aktuell beste Tour zusaetzlich eingefuegt:
        population.insert(int(len(population)/3),self.bestesIndividuum(population))
        akku = []
        while gesamtzahl > len(akku):
            pos1,pos2 = waehlePositionen(len(population))
            akku.append(self.selektiere(population[pos1],population[pos2]))
        return akku

    def bewertung(self, population):
        # Führt die Bewertung der gesamten Population durch. Dazu wird die 
        # Populationsliste in eine Liste mit Paaren [Individuum Bewertung]
        # verwandelt.
        akku = []
        for i in range(len(population)):
            akku.append([population[i],self.bewerte(population[i])])
        return akku

    def bestesIndividuum(self, population):
        # Für die grafische Darstellung, aber auch für das Ergebnis des GA
        # ist es sinnvoll, das 'beste' Individuum bestimmen zu können.
        indexBestes = 0
        for i in range(1,len(population)):
            if besserP(population[i],population[indexBestes]):
                indexBestes = i
        return population[indexBestes]

    def gibSchrittDaten(self):
        return self.schrittDaten
    
    def sichereBeste(self,tour):
        # sichert die Daten der jeweils besten Touren
        wert=bewerte(tour)
        self.schrittDaten.append([tour,wert])

    def gibBeste(self):
        if self.schrittDaten==[]: return []
        return self.schrittDaten[len(self.schrittDaten)-1]
        
    def loescheBeste(self):
        self.schrittDaten=[]

    def ausgabe(self, individuum):
        # Fuer die Ausgabe der ('besten') Loesungen in der Shell
        print(individuum," --> ",bewerte(individuum))

    def genetischerAlgorithmus(self,
                               populationsGroesse,
                               anzahlSchritte,
                               mutationsWahrscheinlichkeit,
                               crossingOverWahrscheinlichkeit):
        "Der eigentliche Prozess"
        population = self.erzeugePopulation(self.staedteListe,populationsGroesse)
        self.loescheBeste()
        self.ausgabe(self.bestesIndividuum(population))
        for i in range(anzahlSchritte):
            population= self.selektion(self.kreuzen(self.mutation(population,
                                                                  mutationsWahrscheinlichkeit),
                                                    crossingOverWahrscheinlichkeit),
                                       populationsGroesse)
            self.sichereBeste(self.bestesIndividuum(population))
        beste=self.bestesIndividuum(population)
        return [beste,bewerte(beste)]

    def gA_Neu(self,
               anzahl,
               populationsGroesse,
               anzahlSchritte,
               mutationsWahrscheinlichkeit,
               crossingOverWahrscheinlichkeit):
        "Aufruf, der die Staedteliste neu erzeugt"
        self.setzeAnzahlStaedte(anzahl)
        self.erzeugeAlleStaedte(anzahl)
        beste=self.genetischerAlgorithmus(populationsGroesse,
                                          anzahlSchritte,
                                          mutationsWahrscheinlichkeit,
                                          crossingOverWahrscheinlichkeit)
        return beste

### Tests
##tspga = TSPGA()
##tspga.genetischerAlgorithmus(20,10,20,20)
##print(len(tspga.gibStaedte()))
##print(len(tspga.gibSchrittDaten()))
##tspga.gibBeste()
##tspga.gA_Neu(3,20,10,20,20)
##print(len(tspga.gibStaedte()))
##print(len(tspga.gibSchrittDaten()))
##tspga.gibBeste()
